#!/bin/sh
# @(#) $Id: ctags.sh 1524 2019-07-09 19:07:03Z leres $ (LBL)

prog=`basename $0`

t1=_${prog}.$$

usage()
{
	echo "usage: ${prog} [-wtd] FILE ..." 1>&2
	exit 1
}

filter()
{
	awk '
	BEGIN {
		skip_n_acl_h["stats"] = 1
		skip_n_acld_h["acl"] = 1
		skip_n_acld_h["req"] = 2
		skip_n_acld_h["rereq"] = 1
		skip_n_acld_h["state"] = 1
	}
	$1 ~ /^(addr|cf|state)$/ && $2 !~ /\.h$/ {
		# Skip any of these when not in an include
		next
	}
	$2 == "acl.h" {
		# Skip the first N occurances
		if (skip_n_acl_h[$1] > 0) {
			--skip_n_acl_h[$1]
			next
		}
	}
	$2 == "acld.h" {
		# Skip the first N occurances
		if (skip_n_acld_h[$1] > 0) {
			--skip_n_acld_h[$1]
			next
		}
	}
	{
		print
	}
	'
}

trap 'rm -f tags ${t1}; exit 1' 1 2 3 15

args=`getopt wtd $*`
if [ $? -ne 0 ]; then
        usage
fi

set -- ${args}

flags=""
for a do
        case "${a}" in
 
        -w)
		flags="${flags}w"
                shift
                ;;
 
        -t)
		flags="${flags}t"
                shift
                ;;
 
        -d)
		flags="${flags}d"
                shift
                ;;

        --)
                shift
                break
                ;;
        esac
done

if [ $# -eq 0 ]; then
        usage
fi

if [ -n "${flags}" ]; then
	flags="-${flags}"
fi

ctags ${flags} $*
mv tags ${t1}
filter < ${t1} > tags
rm ${t1}
