/* @(#) $Id: replicant.h 1507 2019-06-12 01:39:13Z leres $ (LBL) */
#ifndef acld_replicant_h
#define acld_replicant_h

/* Replicant request states */
enum rereqstate {
	RESTATE_PENDING = 0,	/* not processed yet */
	RESTATE_REPLICANT,	/* waiting for replicant to reply */
	RESTATE_READRESPONSE,	/* reading replicant response */
	RESTATE_DONE		/* ok to remove from queue */
};

/* Request flags */
enum rereqflags {
	REFLAG_FAILED	= 0x2,	/* request resulted in a failed status */
};

/* Replicant request */
struct rereq {
	struct rereq *next;
	struct req *req;		/* link to client request */
	enum reqtype type;		/* REQ_* */
	enum rereqstate state;		/* RESTATE_* */
	enum rereqflags reflags;	/* replicant request flags */
	int retries;
	const char *cmd;		/* command name */
	char *aclname;			/* listacl, compact, drop udp/tcp ACL */
	char *prefixname;		/* prefix, noprefix and listprefix */
	u_int32_t cookie;
	char *comment;			/* comment sent to replicant */
	char *response;			/* response received from replicant */
	struct aroute nullzero;
	struct addr prefix;
	struct addr whitelist;
	struct acl acl;
};

extern void replicantappendreq(struct rereq *, struct rereq **);
extern void replicantbump(struct state *);
extern void replicantdone(struct state *, struct rereq *);
extern void replicantfreerereq(struct state *, struct rereq *);
extern void replicantinit(struct state *);
extern void replicantinput(struct state *);
extern void replicantlogin(struct state *);
extern void replicantprocess(struct state *);
extern void replicantreqhook(struct state *, struct req *);
extern void replicantsend(struct state *, struct rereq *);
extern void replicantterminate(struct state *, const char *);
extern void replicantwrite(struct state *, const char *, ...);
#endif
